"""
# Usage
- Set `CONNECT_WITH_STUDY_SERVER` in overcooked_gui.py to True.
- Run this script. Copy the manager id that is printed
- Run the game_server.py script with the manager id copied from the terminal
```
python game_server.py --manager_ids COPIED_UUID
```
- Run 2 overcooked_gui.py scripts in different terminals. For more players change `NUMBER_PLAYER_PER_ENV` and start more guis.

The environment starts when all players connected.
"""

import argparse
import asyncio
import logging
from typing import Tuple

import requests
import uvicorn
from fastapi import FastAPI

from overcooked_simulator import ROOT_DIR
from overcooked_simulator.game_server import CreateEnvironmentConfig
from overcooked_simulator.server_results import PlayerInfo
from overcooked_simulator.utils import (
    url_and_port_arguments,
    add_list_of_manager_ids_arguments,
)

NUMBER_PLAYER_PER_ENV = 2

log = logging.getLogger(__name__)


app = FastAPI()

game_server_url = "localhost:8000"
server_manager_id = None


# @app.get("/")
# async def root(response_class=HTMLResponse):
#     return """
#     <html>
#         <head>
#             <title>Overcooked Game</title>
#         </head>
#         <body>
#             <h1>Start Game!</h1>
#             <button type="button">Click Me!</button>
#         </body>
#     </html>
#     """

running_envs: dict[str, Tuple[int, dict[str, PlayerInfo], list[str]]] = {}
current_free_envs = []


@app.post("/connect_to_game/{request_id}")
async def want_to_play(request_id: str):
    global current_free_envs
    # TODO based on study desing / internal state of request id current state (which level to play)
    if current_free_envs:
        current_free_env = current_free_envs.pop()

        running_envs[current_free_env][2].append(request_id)
        new_running_env = (
            running_envs[current_free_env][0] + 1,
            running_envs[current_free_env][1],
            running_envs[current_free_env][2],
        )
        player_info = running_envs[current_free_env][1][str(new_running_env[0])]
        running_envs[current_free_env] = new_running_env
        if new_running_env[0] < NUMBER_PLAYER_PER_ENV - 1:
            current_free_env.append(current_free_env)
        return player_info
    else:
        environment_config_path = ROOT_DIR / "game_content" / "environment_config.yaml"
        layout_path = ROOT_DIR / "game_content" / "layouts" / "basic.layout"
        item_info_path = ROOT_DIR / "game_content" / "item_info.yaml"
        with open(item_info_path, "r") as file:
            item_info = file.read()
        with open(layout_path, "r") as file:
            layout = file.read()
        with open(environment_config_path, "r") as file:
            environment_config = file.read()
        creation_json = CreateEnvironmentConfig(
            manager_id=server_manager_id,
            number_players=NUMBER_PLAYER_PER_ENV,
            environment_settings={"all_player_can_pause_game": False},
            item_info_config=item_info,
            environment_config=environment_config,
            layout_config=layout,
            seed=1234567890,
        ).model_dump(mode="json")
        # todo async
        env_info = requests.post(
            game_server_url + "/manage/create_env/", json=creation_json
        )

        if env_info.status_code == 403:
            raise ValueError(f"Forbidden Request: {env_info.json()['detail']}")
        env_info = env_info.json()
        print(env_info)
        running_envs[env_info["env_id"]] = (0, env_info["player_info"], [request_id])
        current_free_envs.append(env_info["env_id"])
        return env_info["player_info"]["0"]


def main(host, port, game_server_url_, manager_id):
    global game_server_url, server_manager_id
    game_server_url = "http://" + game_server_url_
    server_manager_id = manager_id[0]
    print(f"Use {server_manager_id=} for {game_server_url=}")
    loop = asyncio.new_event_loop()
    config = uvicorn.Config(app, host=host, port=port, loop=loop)
    server = uvicorn.Server(config)
    loop.run_until_complete(server.serve())


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="Overcooked Simulator Study Server",
        description="Study Server: Match Making, client pre and post managing.",
        epilog="For further information, see https://scs.pages.ub.uni-bielefeld.de/cocosy/overcooked-simulator/overcooked_simulator.html",
    )
    url_and_port_arguments(parser=parser, server_name="Study Server", default_port=8080)
    add_list_of_manager_ids_arguments(parser=parser)
    args = parser.parse_args()
    main(
        args.url,
        args.port,
        game_server_url_="localhost:8000",
        manager_id=args.manager_ids,
    )
